package cc.fascinated.bat.service;

import cc.fascinated.bat.BatApplication;
import cc.fascinated.bat.common.*;
import cc.fascinated.bat.exception.BadRequestException;
import cc.fascinated.bat.exception.ResourceNotFoundException;
import cc.fascinated.bat.model.beatsaber.scoresaber.*;
import cc.fascinated.bat.model.guild.BatGuild;
import cc.fascinated.bat.model.guild.profiles.ScoreSaberScoreFeedProfile;
import cc.fascinated.bat.model.user.profiles.ScoreSaberProfile;
import com.google.gson.JsonObject;
import lombok.NonNull;
import lombok.SneakyThrows;
import lombok.extern.log4j.Log4j2;
import net.dv8tion.jda.api.EmbedBuilder;
import net.dv8tion.jda.api.entities.Guild;
import net.dv8tion.jda.api.entities.MessageEmbed;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.web.socket.CloseStatus;
import org.springframework.web.socket.TextMessage;
import org.springframework.web.socket.WebSocketSession;
import org.springframework.web.socket.client.standard.StandardWebSocketClient;
import org.springframework.web.socket.handler.TextWebSocketHandler;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;

@Service @Log4j2(topic = "ScoreSaber Service")
public class ScoreSaberService extends TextWebSocketHandler {
    private final GuildService guildService;

    private static final String SCORESABER_API = "https://scoresaber.com/api/";
    private static final String GET_PLAYER_ENDPOINT = SCORESABER_API + "player/%s/full";
    private static final String GET_PLAYER_SCORES_ENDPOINT = SCORESABER_API + "player/%s/scores?limit=100&sort=%s&page=%s&withMetadata=true";

    @Autowired
    public ScoreSaberService(@NonNull GuildService guildService) {
        this.guildService = guildService;
        connectWebSocket();
    }

    /**
     * Gets the account from the ScoreSaber API.
     *
     * @param id The id of the account.
     * @return The account.
     * @throws ResourceNotFoundException If the account is not found.
     * @throws cc.fascinated.bat.exception.RateLimitException If the ScoreSaber rate limit is reached.
     */
    public ScoreSaberAccountToken getAccount(String id) {
        ScoreSaberAccountToken account = WebRequest.getAsEntity(String.format(GET_PLAYER_ENDPOINT, id), ScoreSaberAccountToken.class);
        if (account == null) { // Check if the account doesn't exist.
            log.info("Account with id '{}' not found.", id);
            throw new ResourceNotFoundException("Account with id '%s' not found.".formatted(id));
        }
        if (account.isBanned()) {
            throw new BadRequestException("Account with id '%s' is banned.".formatted(id));
        }
        if (account.isInactive()) {
            throw new BadRequestException("Account with id '%s' is inactive.".formatted(id));
        }
        return account;
    }

    /**
     * Gets the scores for the account.
     *
     * @param profile The profile.
     * @param page The page to get the scores from.
     * @return The scores.
     */
    public ScoreSaberScoresPageToken getPageScores(ScoreSaberProfile profile, int page) {
        log.info("Fetching scores for account '{}' from page {}.", profile.getId(), page);
        ScoreSaberScoresPageToken pageToken = WebRequest.getAsEntity(String.format(GET_PLAYER_SCORES_ENDPOINT, profile.getId(), "recent", page), ScoreSaberScoresPageToken.class);
        if (pageToken == null) { // Check if the page doesn't exist.
            return null;
        }
        // Sort the scores by newest time set.
        pageToken.setPlayerScores(Arrays.stream(pageToken.getPlayerScores())
                .sorted((a, b) -> DateUtils.getDateFromString(b.getScore().getTimeSet()).compareTo(DateUtils.getDateFromString(a.getScore().getTimeSet())))
                .toArray(ScoreSaberPlayerScoreToken[]::new));
        return pageToken;
    }

    /**
     * Gets the scores for the account.
     *
     * @param profile The profile.
     * @return The scores.
     */
    public List<ScoreSaberScoresPageToken> getScores(ScoreSaberProfile profile) {
        List<ScoreSaberScoresPageToken> scores = new ArrayList<>(List.of(getPageScores(profile, 1)));
        ScoreSaberPageMetadataToken metadata = scores.get(0).getMetadata();
        int totalPages = (int) Math.ceil((double) metadata.getTotal() / metadata.getItemsPerPage());
        log.info("Fetching {} pages of scores for account '{}'.", totalPages, profile.getId());
        for (int i = 2; i <= totalPages; i++) {
            scores.add(getPageScores(profile, i));
        }

        return scores;
    }

    /**
     * Builds an embed for a score.
     *
     * @param score The score.
     * @return The embed.
     */
    public MessageEmbed buildScoreEmbed(ScoreSaberPlayerScoreToken score) {
        ScoreSaberScoreToken scoreToken = score.getScore();
        ScoreSaberLeaderboardToken leaderboardToken = score.getLeaderboard();
        ScoreSaberScoreToken.LeaderboardPlayerInfo playerInfo = scoreToken.getLeaderboardPlayerInfo();
        return new EmbedBuilder()
                .setAuthor(playerInfo.getName() + " just set a new score!", "https://scoresaber.com/u/%s".formatted(playerInfo.getId()),
                        "https://cdn.scoresaber.com/avatars/%s.jpg".formatted(playerInfo.getId()))
                .setDescription("**%s** (%s%s)\n[[Map Link]](%s) [[SS Profile]](%s)".formatted(
                        leaderboardToken.getSongName(),
                        ScoreSaberUtils.getFormattedDifficulty(leaderboardToken.getDifficulty().getDifficulty()),
                        leaderboardToken.isRanked() ? " " + leaderboardToken.getStars() + "⭐" : "",
                        "https://scoresaber.com/leaderboard/%s".formatted(leaderboardToken.getId()),
                        "https://scoresaber.com/u/%s".formatted(playerInfo.getId())
                ))
                .addField("Accuracy", "%s%%".formatted(
                        leaderboardToken.getMaxScore() == 0 ? "N/A" : NumberUtils.formatNumberCommas(((double) scoreToken.getBaseScore() / leaderboardToken.getMaxScore()) * 100)
                ), true)
                .addField("Raw PP", scoreToken.getPp() == 0 ? "Unranked" : NumberUtils.formatNumberCommas(scoreToken.getPp()), true)
                .addField("Global Rank", "#%s".formatted(NumberUtils.formatNumberCommas(scoreToken.getRank())), true)
                .addField("Misses", "%s".formatted(scoreToken.getMissedNotes()), true)
                .addField("Bad Cuts", "%s".formatted(scoreToken.getBadCuts()), true)
                .addField("Max Combo", "%s %s".formatted(
                        scoreToken.getMaxCombo(),
                        scoreToken.getMaxCombo() == leaderboardToken.getMaxScore() ? "(FC)" : ""
                ), true)
                .setColor(Colors.DEFAULT)
                .setTimestamp(DateUtils.getDateFromString(scoreToken.getTimeSet()).toInstant())
                .build();
    }

    /**
     * Connects to the ScoreSaber WebSocket.
     */
    @SneakyThrows
    private void connectWebSocket() {
        log.info("Connecting to the ScoreSaber WSS.");
        new StandardWebSocketClient().execute(this, "wss://scoresaber.com/ws").get();
    }

    @Override
    public void afterConnectionEstablished(@NonNull WebSocketSession session) {
        log.info("Connected to the ScoreSaber WSS.");
    }

    @Override
    public void afterConnectionClosed(@NonNull WebSocketSession session, @NonNull CloseStatus status) {
        log.info("Disconnected from the ScoreSaber WSS.");
        connectWebSocket(); // Reconnect to the WebSocket.
    }

    @Override @SneakyThrows
    protected void handleTextMessage(@NonNull WebSocketSession session, @NonNull TextMessage message) {
        // Ignore the connection message.
        if (message.getPayload().equals("Connected to the ScoreSaber WSS")) {
            return;
        }

        try {
            JsonObject json = BatApplication.GSON.fromJson(message.getPayload(), JsonObject.class);
            String command = json.get("commandName").getAsString();
            JsonObject data = json.get("commandData").getAsJsonObject();

            if (command.equals("score")) {
                ScoreSaberPlayerScoreToken score = BatApplication.GSON.fromJson(data, ScoreSaberPlayerScoreToken.class);
                ScoreSaberScoreToken.LeaderboardPlayerInfo playerInfo = score.getScore().getLeaderboardPlayerInfo();

                for (Guild guild : DiscordService.JDA.getGuilds()) {
                    BatGuild batGuild = guildService.getGuild(guild.getId());
                    if (batGuild == null) {
                        continue;
                    }

                    ScoreSaberScoreFeedProfile profile = batGuild.getProfile(ScoreSaberScoreFeedProfile.class);
                    if (profile == null) {
                        continue;
                    }
                    if (profile.getChannelId() == null) {
                        continue;
                    }
                    if (!profile.getTrackedUsers().contains(playerInfo.getId())) {
                        continue;
                    }

                    profile.getAsTextChannel().sendMessageEmbeds(this.buildScoreEmbed(score)).queue();
                }
            }
        } catch (Exception ex) {
            log.error("An error occurred while handling the message.", ex);
        }
    }
}
